functor Main(
	 structure Globals: GLOBALS
	 structure FileUtils: FILEUTILS
	 structure Parse: PARSE
	 structure TopLevel: TOPLEVEL
	 val systemName : string
	 val buildtime : string
	 val version : string
	 val comments : string
	 val profiling : bool
	 sharing TopLevel.ParseRes = Parse.ParseRes
	) = struct

(* The top level of a running system.  

   The work of implementing the top-level read-eval-print loop is
   divided into language-independent parts (parsing command-line
   arguments, dealing with files, ...), which are implemented here, 
   and language-dependent parts, which are implemented in the structure 
   TopLevel. *)

open Globals;
open Parse.ParseRes;

val TERSE = createflag ["TERSE","terse"] false

val input_from_file = ref "";
val _ = resetproc (fn()=> input_from_file := "");
val output_to_file = ref "";
val _ = resetproc (fn()=> output_to_file := "");
val tee_output = createflag ["tee_output"] false;

fun besilent _ = (push_stdwr (Wr.to_nowhere()));
fun beloud _ = (pop_stdwr());
val _ = register_cmd "besilent" besilent;
val _ = register_cmd "beloud" beloud;

(* Run a command silently, but turn printing back on before 
   trapping an error.  (This doesn't actually work, since errors
   are trapped and reported at a lower level than the point
   where silently is called below!) *)
fun silently f =
  (besilent();
   (f(); beloud())
   handle s => (beloud(); raise s))

fun twocolumn _ = (setmargin 51)
val _ = register_cmd "twocolumn" twocolumn;

val signal_error_on_exit = createflag ["signal_error"] false;
val _ = register_cmd "error" (fn _ => signal_error_on_exit := true);
val goto_SML_when_finished = createflag ["gotoSML"] false;
val _ = register_cmd "sml" (fn _ => goto_SML_when_finished := true);

val bailout_on_error = ref(false);
exception LoadFailed;

val printfinalstats = createflag ["printstats"] false

val halt_on_error = createflag ["haltonerror"] true

(* Establish the pervasive context *)
val perv_tld = ref (TopLevel.init());

fun export_my_tld name tld = let
  val e = case name of
               NONE => "export();"
             | SOME(name) => "exportto(\""^name^"\");"
  val s = open_string e
  in 
     perv_tld := tld;
     System.Compile.use_stream s;
     close_in s
  end

val quit = ref false

fun rep_loop done parser error init_tld = let
  fun do_it i tld = 
    (case i of
         Specific(s) => 
	   TopLevel.do_command tld 
	                       (fn result => (result))
	                       error s
       | Use(f) => 
           rep_loop_on_file f tld
       | Nothing => 
           tld
       | Set(name,v) => 
           (set_flag name (string_to_bool v);
	    tld)
       | Cmd(name,v) => 
           (* Handle some special commands magically *)
           (case name of
                "export" => (export_my_tld v tld; tld)
              | "reload" => rep_loop_on_file "perv" (TopLevel.init())
           (* Handle the rest of them dynamically *)
              | _ => (do_cmd name v; tld))
    )
    handle 
      NotImplemented(s) =>
          (outputc std_err ("Not implemented: " ^ s ^ "\n");
           error tld)
    | unknown => 
          (outputc std_err ("Exception: "^(System.exn_name unknown)^"\n"); 
       	   error tld)
  fun do_all [] tld      = tld
    | do_all (i::is) tld = do_all is (do_it i tld)
  in if (done()) 
       then init_tld
       else (let
         val ITEMS(items) = parser()
         val result = do_all items init_tld
         in rep_loop done parser error result
         end)
         handle ParseError => (error init_tld)
  end

and rep_loop_on_file fname tld = let 
      val (dev,real_name) = FileUtils.search_for_file fname
      fun parser() = Parse.stream_parse dev
      fun done() = (!quit) orelse (end_of_stream dev)
      fun error tld = (if !bailout_on_error
                         then raise LoadFailed
                       else if !halt_on_error 
                         then (quit := true; signal_error_on_exit := true)
                       else (); 
                       tld)
      val _ = (quit := false)
      val _ = Parse.begin_parser_context fname;
      val _ = if not (!TERSE) 
                then write ("Reading from \"" ^ real_name ^ "\"...\n\n") 
                else ()
      val result = rep_loop done parser error tld
      val _ = if not (!TERSE)
                then write ("\nClosing " ^ real_name ^ "\n") else ();
      val _ = close_in dev;
      val _ = Parse.end_parser_context();
      in result
      end

(* $$$ figure out the new profiling stuff
val _ =
  if profiling then 
    (statsproc ["profile"] (fn()=> (System.Control.Profile.report std_err));
     resetproc (fn()=> (System.Control.Profile.reset())))
  else ();
*)

fun interact f = let 
  val old_SMLout = !System.Print.out;
  in
    System.Print.out := open_out "/dev/null";
    (* $$$
    if profiling
      then System.Control.Profile.profileOn()
      else();
    *)
    f (!perv_tld);
    if !printfinalstats 
      then allstats()
    (* else $$$
            if profiling
      then (System.Control.Profile.profileOff();
	    System.Control.Profile.report (open_out "profile.tmp")) *)
    else ();
    close_wrs();
    System.Print.out := old_SMLout
  end

fun top tld = let 
  fun top_done () = (output (std_out, (!Parse.prompt_char) ^ " "); flush_out std_out;
                     end_of_stream(std_in))
  fun top_error tld = go tld
  and go tld = rep_loop top_done Parse.top_parse top_error tld
  in
    write (systemName ^ "\n");
    write "Copyright (C) 1992, 1993 Benjamin C. Pierce\n";
    write ("Version " ^ version ^ " of " ^ buildtime ^ "\n");
    write "\n";
    write "This program comes with absolutely no warranty.  It is free\n";
    write "software, and you are welcome to redistribute it under certain\n";
    write "conditions; see the file LICENSE for more details.\n";
    write "\n";
    if comments<>"" then (write comments; write "\n") else ();
    Parse.begin_parser_context "";
    go tld;
    Parse.end_parser_context
  end

fun parse_switches ("-i"::s::rest) 
 	= (input_from_file := s;
 	   parse_switches rest)
  | parse_switches ("-o"::s::rest) 
 	= (output_to_file := s;
 	   parse_switches rest)
  | parse_switches ("-tee"::s::rest) 
 	= (output_to_file := s;
 	   tee_output := true;
 	   parse_switches rest)
  | parse_switches ("-set"::f::rest) 
 	= (set_flag f true;
 	   parse_switches rest)
  | parse_switches ("-reset"::f::rest) 
 	= (set_flag f false;
 	   parse_switches rest)
  | parse_switches ("-latex"::rest) 
 	= (set_flag "TERSE" true;
           set_flag "latex" true;
           parse_switches rest)
  | parse_switches ("-sml"::rest) 
 	= (goto_SML_when_finished := true;
 	   parse_switches rest)
  | parse_switches (s::rest) 
 	= (input_from_file := s;
 	   parse_switches rest)
  | parse_switches ([]) 
  	= ()

fun rep_command_line(argv,env) = 
  (global_reset();
   (* load local commands or extensions to the pervasive environment, 
      contained in an (optional) file "local" *)
   ((silently(fn()=> (
        bailout_on_error := true;
        perv_tld := rep_loop_on_file "local" (!perv_tld);
        bailout_on_error := false)))
      handle Io(_) => ()
           | LoadFailed => (output(std_err, "Error occurred while processing local.f\n");
                            raise LoadFailed));
   parse_switches (tl argv);
   if (!output_to_file) = ""
     then ()
     else (set_stdwr (Wr.to_file (!output_to_file))
           handle Io(s) => outputc std_err ("Can't write to " ^ s ^ "\n"));
   if (!input_from_file) = ""
      then (if !goto_SML_when_finished 
              then ()
              else interact top)
      else interact (fn tld => rep_loop_on_file (!input_from_file) tld);
   global_finalize();
   if (!goto_SML_when_finished)
     then (outputc std_err "Escaping to SML top-level\n\n")
     else (System.Unsafe.CInterface.exit 
            (if !signal_error_on_exit then 1 else 0)))

fun process_file s = rep_command_line (["","-sml",s],nil);

fun go () = 
  (global_reset();
   interact top;
   global_finalize())

(* Establish the pervasive context *)
val _ = global_reset();
val _ = perv_tld := rep_loop_on_file "perv" (TopLevel.init())
val _ = global_finalize();

end
